"""Subsumptions visitor."""
from functools import singledispatch
from typing import Set

from pylogics.syntax.base import Formula

from nl2ltl.declare.base import Template
from nl2ltl.declare.declare import (
    Absence,
    ChainResponse,
    Existence,
    ExistenceTwo,
    NotCoExistence,
    Precedence,
    RespondedExistence,
    Response,
)


@singledispatch
def subsumptions(formula: Formula) -> Set[Template]:
    """Compute subsumptions for a Formula."""
    raise NotImplementedError("subsumption handler not implemented for formula %s" % type(formula))


@subsumptions.register
def subsumptions_existence(_formula: Existence) -> Set[Template]:
    """Compute subsumptions for Existence."""
    return set()


@subsumptions.register
def subsumptions_existence_two(formula: ExistenceTwo) -> Set[Template]:
    """Compute subsumptions for ExistenceTwo."""
    return {Existence(formula.argument)}


@subsumptions.register
def subsumptions_absence(_formula: Absence) -> Set[Template]:
    """Compute subsumptions for Absence."""
    return set()


@subsumptions.register
def subsumptions_responded_existence(formula: RespondedExistence) -> Set[Template]:
    """Compute subsumptions for RespondedExistence."""
    return {Existence(formula.operands[0]), Existence(formula.operands[1])}


@subsumptions.register
def subsumptions_response(formula: Response) -> Set[Template]:
    """Compute subsumptions for Response."""
    return {
        Existence(formula.operands[0]),
        Existence(formula.operands[1]),
        RespondedExistence(*formula.operands),
        ChainResponse(*formula.operands),
    }


@subsumptions.register
def subsumptions_precedence(formula: Precedence) -> Set[Template]:
    """Compute subsumptions for Precedence."""
    return {Existence(formula.operands[0]), Existence(formula.operands[1])}


@subsumptions.register
def subsumptions_chain_response(formula: ChainResponse) -> Set[Template]:
    """Compute subsumptions for ChainResponse."""
    return {
        Existence(formula.operands[0]),
        Existence(formula.operands[1]),
        RespondedExistence(*formula.operands),
        ChainResponse(*formula.operands),
    }


@subsumptions.register
def subsumptions_not_co_existence(formula: NotCoExistence) -> Set[Template]:
    """Compute subsumptions for NotCoExistence."""
    return {Existence(formula.operands[0]), Existence(formula.operands[1])}



"""Grounding visitor."""
import logging
from typing import Dict, Set

from pylogics.syntax.ltl import Atomic

from nl2ltl.declare.base import Template
from nl2ltl.declare.declare import (
    Absence,
    ChainResponse,
    Existence,
    ExistenceTwo,
    NotCoExistence,
    Precedence,
    RespondedExistence,
    Response,
)
from nl2ltl.engines.utils import decapitalize


def ground_existence(connectors: Dict[str, float]) -> Set[Template]:
    """Compute ground for Existence."""
    if len(list(connectors)) > 0:
        return {Existence(Atomic(decapitalize(list(connectors)[0])))}
    else:
        logging.warning("No valid matching, cannot instantiate Existence with < 1 connectors.")
        return set()


def ground_existencetwo(connectors: Dict[str, float]) -> Set[Template]:
    """Compute ground for ExistenceTwo."""
    if len(list(connectors)) > 0:
        return {ExistenceTwo(Atomic(decapitalize(list(connectors)[0])))}
    else:
        logging.warning("No valid matching, cannot instantiate ExistenceTwo with < 1 connectors.")
        return set()


def ground_absence(connectors: Dict[str, float]) -> Set[Template]:
    """Compute ground for Absence."""
    if len(list(connectors)) > 0:
        return {Absence(Atomic(decapitalize(list(connectors)[0])))}
    else:
        logging.warning("No valid matching, cannot instantiate Absence with < 1 connectors.")
        return set()


def ground_respondedexistence(connectors: Dict[str, float]) -> Set[Template]:
    """Compute ground for RespondedExistence."""
    if len(list(connectors)) >= 2:
        return {
            RespondedExistence(
                Atomic(decapitalize(list(connectors)[0])),
                Atomic(decapitalize(list(connectors)[1])),
            )
        }
    else:
        logging.warning("No valid matching, cannot instantiate RespondedExistence with < 2 connectors.")
        return set()


def ground_response(connectors: Dict[str, float]) -> Set[Template]:
    """Compute ground for Response."""
    if len(list(connectors)) >= 2:
        return {
            Response(
                Atomic(decapitalize(list(connectors)[0])),
                Atomic(decapitalize(list(connectors)[1])),
            )
        }
    else:
        logging.warning("No valid matching, cannot instantiate Response with < 2 connectors.")
        return set()


def ground_precedence(connectors: Dict[str, float]) -> Set[Template]:
    """Compute ground for Precedence."""
    if len(list(connectors)) >= 2:
        return {
            Precedence(
                Atomic(decapitalize(list(connectors)[0])),
                Atomic(decapitalize(list(connectors)[1])),
            )
        }
    else:
        logging.warning("No valid matching, cannot instantiate Precedence with < 2 connectors.")
        return set()


def ground_chainresponse(connectors: Dict[str, float]) -> Set[Template]:
    """Compute ground for ChainResponse."""
    if len(list(connectors)) >= 2:
        return {
            ChainResponse(
                Atomic(decapitalize(list(connectors)[0])),
                Atomic(decapitalize(list(connectors)[1])),
            )
        }
    else:
        logging.warning("No valid matching, cannot instantiate ChainResponse with < 2 connectors.")
        return set()


def ground_notcoexistence(connectors: Dict[str, float]) -> Set[Template]:
    """Compute ground for NotCoExistence."""
    if len(list(connectors)) >= 2:
        return {
            NotCoExistence(
                Atomic(decapitalize(list(connectors)[1])),
                Atomic(decapitalize(list(connectors)[1])),
            )
        }
    else:
        logging.warning("No valid matching, cannot instantiate NotCoExistence with < 2 connectors.")
        return set()



"""Implementation of the RASA engine.

Repository:

    https://github.com/RasaHQ/rasa

"""
import asyncio
import shutil
from pathlib import Path
from typing import Dict

from pylogics.syntax.base import Formula

from nl2ltl.helpers import requires_optional

try:
    import rasa
    from rasa.core.agent import Agent
except ImportError:
    rasa = Agent = None

from nl2ltl.engines.base import Engine
from nl2ltl.engines.rasa import ENGINE_ROOT
from nl2ltl.engines.rasa.helpers import _get_latest_model
from nl2ltl.engines.rasa.output import RasaOutput, parse_rasa_output, parse_rasa_result
from nl2ltl.filters.base import Filter

engine_root = ENGINE_ROOT
DATA_DIR = engine_root / "data"
DOMAIN_PATH = engine_root / DATA_DIR / "domain.yml"
CONFIG_PATH = engine_root / DATA_DIR / "config.yml"
TRAINING_PATH = engine_root / DATA_DIR / "nlu_training.yml"
MODEL_OUTPUT_PATH = engine_root / "models"


@requires_optional
class RasaEngine(Engine):
    """The RASA engine."""

    def __init__(
        self,
        model: Path = None,
    ):
        """Rasa NLU Engine initialization."""
        self._load_model(model)

        self._check_consistency()

    def _load_model(self, model: Path = None):
        if model:
            self.agent = Agent.load(model)
        else:
            self.agent = Agent.load(_get_latest_model(path=MODEL_OUTPUT_PATH))

    def _check_consistency(self) -> None:
        """Run consistency checks."""
        self.__check_rasa_available()
        self.__check_rasa_version()

    def __check_rasa_available(self):
        """Check that the Rasa tool is available."""
        is_rasa_present = shutil.which("rasa") is not None
        if is_rasa_present is None:
            raise Exception(
                "Rasa is not installed. Please follow"
                "the installation instructions at https://rasa.com/docs/rasa/installation.\n"
            )

    def __check_rasa_version(self):
        """Check that the Rasa tool is at the right version."""
        is_right_version = rasa.__version__ == "3.6.16"
        if not is_right_version:
            raise Exception(
                "Rasa needs to be at version 3.6.16. "
                "Please install it manually using:"
                "\n"
                "pip install rasa==3.6.16"
            )

    @staticmethod
    def train(
        domain: Path = DOMAIN_PATH,
        config: Path = CONFIG_PATH,
        training: Path = TRAINING_PATH,
    ) -> Path:
        """Train a Rasa model."""
        result = rasa.train(
            domain=str(domain),
            config=str(config),
            training_files=[str(training)],
            output=str(MODEL_OUTPUT_PATH),
            force_training=True,
        )
        return Path(result.model)

    def translate(self, utterance: str, filtering: Filter = None) -> Dict[Formula, float]:
        """From NL to best matching LTL formulas with confidence."""
        return _process_utterance(utterance, self.agent, filtering)


def _process_utterance(utterance: str, rasa_agent: Agent, filtering: Filter) -> Dict[Formula, float]:
    """Process NL utterance.

    :param utterance: the natural language utterance
    :return: a dict with matching formulas and confidence
    """
    prediction = asyncio.run(rasa_agent.parse_message(utterance.strip()))
    rasa_result: RasaOutput = parse_rasa_output(prediction)
    matching_formulas: Dict[Formula, float] = parse_rasa_result(rasa_result, filtering)
    return matching_formulas



"""Base classes for filters."""

from abc import ABC, abstractmethod
from typing import Dict

from pylogics.syntax.base import Formula


class Filter(ABC):
    """Base class for all templates."""

    NAME: str

    @abstractmethod
    def enforce(self, output: Dict[Formula, float], entities: Dict[str, float], **kwargs) -> Dict[Formula, float]:
        """Enforce filters on output."""
        raise NotImplementedError("Filters must implement this method.")



"""Conflicts visitor."""
from functools import singledispatch
from typing import Set

from pylogics.syntax.base import Formula

from nl2ltl.declare.base import Template
from nl2ltl.declare.declare import (
    Absence,
    ChainResponse,
    Existence,
    ExistenceTwo,
    NotCoExistence,
    Precedence,
    RespondedExistence,
    Response,
)


@singledispatch
def conflicts(formula: Formula) -> Set[Template]:
    """Compute conflicts for a Formula."""
    raise NotImplementedError("handler not implemented for formula %s" % type(formula))


@conflicts.register
def conflicts_existence(formula: Existence) -> Set[Template]:
    """Compute conflicts for Existence."""
    return {Absence(formula.argument)}


@conflicts.register
def conflicts_existence_two(formula: ExistenceTwo) -> Set[Template]:
    """Compute conflicts for ExistenceTwo."""
    return {Absence(formula.argument)}


@conflicts.register
def conflicts_absence(formula: Absence) -> Set[Template]:
    """Compute conflicts for Absence."""
    return {Existence(formula.argument)}


@conflicts.register
def conflicts_responded_existence(formula: RespondedExistence) -> Set[Template]:
    """Compute conflicts for RespondedExistence."""
    return {
        Absence(formula.operands[0]),
        Absence(formula.operands[1]),
        Precedence(*reversed(formula.operands)),
    }


@conflicts.register
def conflicts_response(formula: Response) -> Set[Template]:
    """Compute conflicts for Response."""
    return {
        Absence(formula.operands[0]),
        Absence(formula.operands[1]),
        Precedence(*reversed(formula.operands)),
    }


@conflicts.register
def conflicts_precedence(formula: Precedence) -> Set[Template]:
    """Compute conflicts for Precedence."""
    return {
        RespondedExistence(*reversed(formula.operands)),
        Response(*reversed(formula.operands)),
        ChainResponse(*reversed(formula.operands)),
    }


@conflicts.register
def conflicts_chain_response(formula: ChainResponse) -> Set[Template]:
    """Compute conflicts for ChainResponse."""
    return {
        Precedence(*reversed(formula.operands)),
        Absence(formula.operands[0]),
        Absence(formula.operands[1]),
    }


@conflicts.register
def conflicts_not_co_existence(_formula: NotCoExistence) -> Set[Template]:
    """Compute conflicts for NotCoExistence."""
    return set()



"""This module contains utilities to call the Lydia tool from Python."""
import glob
import os
from pathlib import Path


def _get_latest_model(path: Path):
    """Get the Rasa latest model."""
    if not path:
        return None
    list_of_files = glob.glob(os.path.join(path, "*.tar.gz"))
    if len(list_of_files) == 0:
        return None
    return max(list_of_files, key=os.path.getctime)



"""Classes for declare templates."""
from pylogics.syntax.base import And, Formula, Implies, Not, Or, _BinaryOp, _UnaryOp
from pylogics.syntax.ltl import Always, Eventually, Next, Until
from pylogics.syntax.pltl import Before, Historically, Once, Since

from nl2ltl.declare.base import Template, TemplateEnum
from nl2ltl.declare.misc import enforce_binary, enforce_unary


class Existence(Template, _UnaryOp):
    """The existence template."""

    SYMBOL = TemplateEnum.EXISTENCE.value

    def __init__(self, argument):
        """Existence initialization."""
        super().__init__(arg=argument)
        enforce_unary(self.argument)

    def to_ltlf(self) -> Formula:
        """Translate Existence to LTLf."""
        return Eventually(self.argument)

    def to_english(self) -> str:
        """English meaning."""
        return f"Eventually, {self.argument} will happen."

    def to_ppltl(self) -> Formula:
        """Translate Existence to PPLTL."""
        return Once(self.argument)


class ExistenceTwo(Template, _UnaryOp):
    """The existenceTwo template."""

    SYMBOL = TemplateEnum.EXISTENCE_TWO.value

    def __init__(self, argument):
        """Existence Two initialization."""
        super().__init__(arg=argument)
        enforce_unary(self.argument)

    def to_ltlf(self) -> Formula:
        """Translate ExistenceTwo to LTLf."""
        return Eventually(And(self.argument, Next(Eventually(self.argument))))

    def to_english(self) -> str:
        """English meaning."""
        return f"{self.argument} will happen at least twice."

    def to_ppltl(self) -> Formula:
        """Translate ExistenceTwo to PPLTL."""
        return Once(And(self.argument, Before(Once(self.argument))))


class Absence(Template, _UnaryOp):
    """The existenceTwo template."""

    SYMBOL = TemplateEnum.ABSENCE.value

    def __init__(self, argument):
        """Absence initialization."""
        super().__init__(arg=argument)
        enforce_unary(self.argument)

    def to_ltlf(self) -> Formula:
        """Translate Absence to LTLf."""
        return Not(Eventually(self.argument))

    def to_english(self) -> str:
        """English meaning."""
        return f"{self.argument} will never happen."

    def to_ppltl(self) -> Formula:
        """Translate Absence to PPLTL."""
        return Not(Once(self.argument))


class RespondedExistence(Template, _BinaryOp):
    """The RespondedExistence template."""

    SYMBOL = TemplateEnum.RESPONDED_EXISTENCE.value

    def __init__(self, proposition1, proposition2):
        """Responded Existence initialization."""
        super().__init__(proposition1, proposition2)
        enforce_binary(self.operands)

    def to_ltlf(self) -> Formula:
        """Translate RespondedExistence to LTLf."""
        return Implies(Eventually(self.operands[0]), Eventually(self.operands[1]))

    def to_english(self) -> str:
        """English meaning."""
        return (
            f"If {self.operands[0]} happens at least once then {self.operands[1]} has to happen or happened "
            f"before {self.operands[0]}."
        )

    def to_ppltl(self) -> Formula:
        """Translate RespondedExistence to PPLTL."""
        return Implies(Once(self.operands[0]), Once(self.operands[1]))


class Response(Template, _BinaryOp):
    """The Response template."""

    SYMBOL = TemplateEnum.RESPONSE.value

    def __init__(self, proposition1, proposition2):
        """Response initialization."""
        super().__init__(proposition1, proposition2)
        enforce_binary(self.operands)

    def to_ltlf(self) -> Formula:
        """Translate Response to LTLf."""
        return Always(Implies(self.operands[0], Eventually(self.operands[1])))

    def to_english(self) -> str:
        """English meaning."""
        return f"Whenever  {self.operands[0]} happens,  {self.operands[1]} has to happen " f"eventually afterward."

    def to_ppltl(self) -> Formula:
        """Translate Response to PPLTL."""
        return Not(Since(Not(self.operands[1]), And(self.operands[0], Not(self.operands[1]))))


class Precedence(Template, _BinaryOp):
    """The Precedence template."""

    SYMBOL = TemplateEnum.PRECEDENCE.value

    def __init__(self, proposition1, proposition2):
        """Precedence initialization."""
        super().__init__(proposition1, proposition2)
        enforce_binary(self.operands)

    def to_ltlf(self) -> Formula:
        """Translate Precedence to LTLf."""
        return Or(
            Until(Not(self.operands[1]), self.operands[0]),
            Always(Not(self.operands[1])),
        )

    def to_english(self) -> str:
        """English meaning."""
        return f"Whenever  {self.operands[1]} happens,  {self.operands[0]} has to have happened " f"before it."

    def to_ppltl(self) -> Formula:
        """Translate Precedence to PPLTL."""
        return Or(
            Once(
                And(
                    self.operands[0],
                    Historically(Or(self.operands[0], Not(self.operands[1]))),
                )
            ),
            Historically(Not(self.operands[1])),
        )


class ChainResponse(Template, _BinaryOp):
    """The ChainResponse template."""

    SYMBOL = TemplateEnum.CHAIN_RESPONSE.value

    def __init__(self, proposition1, proposition2):
        """Chain Response initialization."""
        super().__init__(proposition1, proposition2)
        enforce_binary(self.operands)

    def to_ltlf(self) -> Formula:
        """Translate ChainResponse to LTLf."""
        return Always(Implies(self.operands[0], Next(self.operands[1])))

    def to_english(self) -> str:
        """English meaning."""
        return (
            f"Every time  {self.operands[0]} happens, it must be directly followed by  "
            f"{self.operands[1]} ( {self.operands[1]} can also follow other activities)."
        )

    def to_ppltl(self) -> Formula:
        """Translate ChainResponse to PPLTL."""
        return And(
            Historically(Implies(Before(self.operands[0]), self.operands[1])),
            Not(self.operands[0]),
        )


class NotCoExistence(Template, _BinaryOp):
    """The NotCoExistence template."""

    SYMBOL = TemplateEnum.NOT_CO_EXISTENCE.value

    def __init__(self, proposition1, proposition2):
        """Not Co Existence initialization."""
        super().__init__(proposition1, proposition2)
        enforce_binary(self.operands)

    def to_ltlf(self) -> Formula:
        """Translate NotCoExistence to LTLf."""
        return Implies(Eventually(self.operands[0]), Not(Eventually(self.operands[1])))

    def to_english(self) -> str:
        """English meaning."""
        return f"Either  {self.operands[0]} or {self.operands[1]} can happen, but not both."

    def to_ppltl(self) -> Formula:
        """Translate NotCoExistence to PPLTL."""
        return Implies(Once(self.operands[0]), Not(Once(self.operands[1])))



"""Greedy filter class."""
from typing import Dict

from pylogics.syntax.base import Formula

from nl2ltl.declare.base import Template
from nl2ltl.filters.base import Filter
from nl2ltl.filters.utils.conflicts import conflicts
from nl2ltl.filters.utils.subsumptions import subsumptions


class BasicFilter(Filter):
    """Basic filter class."""

    NAME = "basic"

    @staticmethod
    def enforce(output: Dict[Formula, float], entities: Dict[str, float], **kwargs) -> Dict[Formula, float]:
        """Enforce conflicts and subsumptions to output formulas.

        Algorithm:
        - scan the output
        - pick the best match formulas based on number of matching entities
        """
        return output


class GreedyFilter(Filter):
    """Greedy filter class."""

    NAME = "greedy"

    @staticmethod
    def enforce(output: Dict[Formula, float], entities: Dict[str, float], **kwargs) -> Dict[Formula, float]:
        """Enforce conflicts and subsumptions to output formulas.

        Algorithm:
        - empty set of formulas
        - add the highest scoring formula to the result set
        - scan the output one formula at a time
        - compare the highest scoring formula with other formulas
        - if the current formulas has conflicts with the highest scoring formula, discard it
        - if the current formula subsumes the highest scoring formula, discard it and keep the highest scoring formula
        - else add the current formula to the result set
        """
        result_set = set()

        highest_scoring_formula = max(output, key=output.get, default=Template)
        formula_conflicts = conflicts(highest_scoring_formula)
        formula_subsumptions = subsumptions(highest_scoring_formula)

        result_set.add(highest_scoring_formula)
        for formula in output:
            if formula in formula_subsumptions:
                continue
            if formula in formula_conflicts:
                continue
            result_set.add(formula)
        return {formula: output[formula] for formula in result_set}



"""Helper functions."""
from typing import Sequence, Type

from pylogics.syntax.base import Formula
from pylogics.syntax.ltl import Atomic


def _enforce(condition: bool, message: str = "", exception_cls: Type[Exception] = AssertionError):
    """User-defined assert."""
    if not condition:
        raise exception_cls(message)
    else:
        return True


def enforce_unary(argument: Formula):
    """Enforce a unary formula."""
    return _enforce(
        isinstance(argument, Atomic),
        "argument is not an instance of 'Atomic'",
        exception_cls=ValueError,
    )


def enforce_binary(operands: Sequence[Formula]):
    """Enforce DECLARE binary condition."""
    return _enforce(
        len(operands) == 2,
        f"expected exactly 2 operands, found {len(operands)} operands",
    ) and _enforce(
        all(map(lambda x: isinstance(x, Atomic), operands)),
        "some argument is not an instance of 'Atomic'",
        exception_cls=ValueError,
    )



"""NL2LTLf core module."""
from typing import Dict

from pylogics.syntax.base import Formula

from nl2ltl.engines import Engine
from nl2ltl.filters.base import Filter


def _call_translation_method(
    utterance: str, engine: Engine, filtering: Filter, method_name: str
) -> Dict[Formula, float]:
    """Call the translation method."""
    method = getattr(engine, method_name)
    return method(utterance, filtering)


def translate(utterance: str, engine: Engine, filtering: Filter = None) -> Dict[Formula, float]:
    """From NL to LTLf.

    :param utterance: the natural language utterance to translate.
    :param engine: the engine to use.
    :param filtering: the filtering function to use.
    :return: the best matching LTL formulas with their confidence.
    """
    return _call_translation_method(utterance, engine, filtering, translate.__name__)



"""This module implements the library exceptions."""
import functools
import sys


def requires_optional(cls):
    @functools.wraps(cls)
    def wrapper_decor(*args, **kwargs):
        if "rasa" not in sys.modules:
            raise ModuleNotFoundError(f"Rasa is required to instantiate {cls.__name__}")
        return cls(*args, **kwargs)

    return wrapper_decor



"""Parse GPT output to produce Dict[Formula, Float] result."""
import re
from dataclasses import dataclass
from typing import Dict, Match, Set, Tuple, cast

from pylogics.syntax.base import Formula

from nl2ltl.engines.utils import _get_formulas
from nl2ltl.filters.base import Filter


@dataclass
class GPTOutput:
    """Dataclass to represent the GPT output."""

    pattern: str
    entities: Tuple[str, ...]

    def __post_init__(self):
        """Do consistency checks after initialization."""
        assert self.pattern is not None
        assert self.entities is not None


@dataclass
class _GPTOutputWrapper:
    """A wrapper to the textual output of GPT."""

    output: dict
    mode: str

    @property
    def pattern(self) -> str:
        """Get the predicted pattern."""
        from nl2ltl.engines.gpt.core import OperationModes

        if self.mode == OperationModes.CHAT.value:
            return str(
                cast(
                    Match,
                    re.search(
                        "PATTERN: (.*)\n",
                        self.output.choices[0].message.content,
                    ),
                ).group(1)
            )
        else:
            return str(
                cast(
                    Match,
                    re.search("PATTERN: (.*)\n", self.output.choices[0].text),
                ).group(1)
            )

    @property
    def entities(self) -> Tuple[str]:
        """Get the predicted entities."""
        from nl2ltl.engines.gpt.core import OperationModes

        if self.mode == OperationModes.CHAT.value:
            return tuple(
                cast(
                    Match,
                    re.search("SYMBOLS: (.*)", self.output.choices[0].message.content),
                )
                .group(1)
                .split(", ")
            )
        else:
            return tuple(cast(Match, re.search("SYMBOLS: (.*)", self.output.choices[0].text)).group(1).split(", "))


def parse_gpt_output(gpt_output: dict, operation_mode: str) -> GPTOutput:
    """Parse the GPT output.

    :param gpt_output: the json description of the GPT prediction.
    :param operation_mode: the operation mode of the GPT engine.
    :return: a GPTOutput instance.
    """
    wrapper = _GPTOutputWrapper(gpt_output, operation_mode)
    pattern: str = wrapper.pattern
    entities: Tuple[str] = wrapper.entities
    gpt_result = GPTOutput(pattern, entities)
    return gpt_result


def parse_gpt_result(output: GPTOutput, filtering: Filter = None) -> Dict[Formula, float]:
    """Build a dict of formulas, given the GPTOutput object.

    :param output: a GPTOutput instance.
    :param filtering: a custom filtering function
    :return: the dictionary of formulas.
    """
    result: Dict[Formula, float] = {}
    symbols: Dict[str, float] = {e: 1 for e in output.entities}
    formulas: Set[Formula] = _get_formulas(output.pattern, symbols)
    if all(isinstance(f, Formula) for f in formulas):
        for f in formulas:
            result[f] = 1
    else:
        raise Exception("The output is not a valid formula.")

    if filtering:
        return filtering.enforce(result, symbols)
    else:
        return result



"""Abstract definition of a engine."""
from abc import ABC, ABCMeta
from typing import Dict

from pylogics.syntax.base import Formula

from nl2ltl.exceptions import NotImplementedEngineFunction
from nl2ltl.filters.base import Filter


class _MetaEngine(ABCMeta):
    """Metaclass for rasa."""

    def __new__(mcs, *args, **kwargs):
        """Instantiate a new class."""
        return super().__new__(mcs, *args, **kwargs)


class Engine(ABC, metaclass=_MetaEngine):
    """Nl2ltlf engine interface."""

    @classmethod
    def __not_supported_error(cls, operation: str) -> Exception:
        """Raise a not supported error."""
        return NotImplementedEngineFunction(f"operation '{operation}' is not supported by the '{cls.__name__}' engine")

    def translate(self, utterance: str, filtering: Filter = None) -> Dict[Formula, float]:
        """Transform a Natural Language utterance into an LTLf formula.

        :param utterance: a Natural Language utterance
        :param filtering: a custom filtering algorithm
        :return: the corresponding LTLf formula template
        """
        raise self.__not_supported_error(self.translate.__name__)



"""Parse Rasa output to produce Dict[Formula, Float] result."""
from dataclasses import dataclass
from typing import Any, Dict

from pylogics.syntax.base import Formula

from nl2ltl.engines.utils import _get_formulas
from nl2ltl.filters.base import Filter


@dataclass
class RasaOutput:
    """Dataclass to represent the Rasa output."""

    text: str
    intent: Dict[str, float]
    entities: Dict[str, float]
    intent_ranking: Dict[str, Any]

    def __post_init__(self):
        """Do consistency checks after initialization."""
        assert self.intent is not None
        assert self.entities is not None


@dataclass
class _RasaOutputWrapper:
    """A wrapper to the textual output of Rasa."""

    output: dict

    @property
    def text(self) -> str:
        """Get the matched text."""
        return str(self.output["text"])

    @property
    def intent(self) -> Dict[str, float]:
        """Get the predicted intent."""
        return {self.output["intent"]["name"]: self.output["intent"]["confidence"]}

    @property
    def entities(self) -> Dict[str, float]:
        """Get the predicted entities."""
        entities: Dict[str, float] = {}
        for item in self.output["entities"]:
            entities[item["value"]] = item["confidence_entity"]
        return entities

    @property
    def intent_ranking(self) -> Dict[str, float]:
        """Get the intent ranking."""
        ranking: Dict[str, float] = {}
        for item in self.output["intent_ranking"]:
            ranking[item["name"]] = item["confidence"]
        return ranking


def parse_rasa_output(rasa_output: dict) -> RasaOutput:
    """Parse the Rasa output.

    :param rasa_output: the json description of the RASA prediction.
    :return: a RasaOutput instance.
    """
    wrapper = _RasaOutputWrapper(rasa_output)
    text: str = wrapper.text
    intent: Dict[str, float] = wrapper.intent
    entities: Dict[str, float] = wrapper.entities
    intent_ranking: Dict[str, float] = wrapper.intent_ranking
    rasa_result = RasaOutput(text, intent, entities, intent_ranking)
    return rasa_result


def parse_rasa_result(output: RasaOutput, filtering: Filter = None) -> Dict[Formula, float]:
    """Build a dict of formulas, given the RasaOutput object.

    :param output: a RasaOutput instance.
    :param filtering: a custom filtering function
    :return: the dictionary of formulas.
    """
    result: Dict[Formula, float] = {}
    for intent_name, confidence in output.intent_ranking.items():
        formulas = _get_formulas(intent_name, output.entities)
        if all(isinstance(f, Formula) for f in formulas):
            for f in formulas:
                result[f] = confidence
    if filtering:
        return filtering.enforce(result, output.entities)
    else:
        return result



"""Engines utils."""
import difflib
from typing import Callable, Dict, Set, Union

from pylogics.syntax.base import AtomName, Formula

from nl2ltl.declare.base import TemplateEnum


def _get_formulas(name: str, args: Dict[str, float]) -> Set[Formula]:
    """Instantiate matching formulas based on intent name and entities."""
    import nl2ltl.declare.declare
    import nl2ltl.engines.grounding

    class_name_match = difflib.get_close_matches(name, [x.value for x in TemplateEnum], n=1)
    grounding_map: Dict[str, Callable] = {}
    for c_name in TemplateEnum:
        grounding_map[c_name.value] = getattr(nl2ltl.engines.grounding, f"ground_{c_name.value.lower()}")

    grounding_func: Callable = grounding_map[str(class_name_match[0])]
    grounded_formulas: Set[Formula] = grounding_func(args)
    return grounded_formulas


def pretty(result: Dict[Formula, float]):
    """Pretty print Rasa output."""
    print("=" * 150)
    for k, v in result.items():
        print(f"Declare Template: {k!s}", end="\n")
        print(f"English meaning:  {k.to_english()}", end="\n")
        print(f"Confidence:       {v!s}", end="\n\n")


def check_(condition: bool, message: str = "") -> None:
    """User-defined assert.

    This function is useful to avoid the use of the built-in assert statement, which is removed
        when the code is compiled in optimized mode. For more information, see
        https://bandit.readthedocs.io/en/1.7.5/plugins/b101_assert_used.html
    """
    if not condition:
        raise AssertionError(message)


def decapitalize(s: Union[str, AtomName]) -> str:
    """Decapitalize a string."""
    return s[:1].lower() + s[1:] if s else ""



"""Tests for Rasa engine."""
from typing import Dict

import pytest

from nl2ltl import translate
from nl2ltl.engines.rasa.core import RasaEngine
from nl2ltl.filters.simple_filters import BasicFilter, GreedyFilter

from .conftest import UtterancesFixtures


class TestRasa:
    """Rasa test class."""

    @classmethod
    def setup_class(cls):
        """Setup any state specific to the execution of the given class (which
        usually contains tests).
        """
        cls.rasa_engine = RasaEngine()
        cls.basic_filter = BasicFilter()
        cls.greedy_filter = GreedyFilter()

    @pytest.mark.parametrize("utterance", UtterancesFixtures.utterances)
    def test_rasa_engine_basic(self, utterance):
        """Test Rasa engine for utterances with basic filter."""
        output = translate(utterance, self.rasa_engine, self.basic_filter)
        assert isinstance(output, Dict)

    @pytest.mark.parametrize("utterance", UtterancesFixtures.utterances)
    def test_rasa_engine_greedy(self, utterance):
        """Test Rasa engine for utterances with greedy filter."""
        output = translate(utterance, self.rasa_engine, self.greedy_filter)
        assert isinstance(output, Dict)



"""Base classes for declare templates."""

from abc import abstractmethod
from enum import Enum, unique

from pylogics.syntax.base import Formula, Logic


@unique
class TemplateEnum(Enum):
    """Template enumeratin class."""

    EXISTENCE = "Existence"
    EXISTENCE_TWO = "ExistenceTwo"
    ABSENCE = "Absence"
    RESPONDED_EXISTENCE = "RespondedExistence"
    RESPONSE = "Response"
    PRECEDENCE = "Precedence"
    CHAIN_RESPONSE = "ChainResponse"
    NOT_CO_EXISTENCE = "NotCoExistence"


class Template(Formula):
    """Base class for all templates."""

    SYMBOL: str

    def logic(self) -> Logic:
        """Get the logic."""
        return Logic.LTL

    @abstractmethod
    def to_ltlf(self) -> "Formula":
        """Get the template translation to LTLf."""
        raise NotImplementedError("Template's subclasses must implement this method.")

    @abstractmethod
    def to_english(self) -> str:
        """Get the English semantics."""
        raise NotImplementedError("Template's subclasses must implement this method.")

    @abstractmethod
    def to_ppltl(self) -> "Formula":
        """Get the template translation to PPLTL."""

    def __hash__(self):
        """Delegate the computation of the hash to the superclass."""
        return super(Formula, self).__hash__()



"""Implementation of the GPT engine.

Website:

    https://openai.com

"""
import json
from enum import Enum
from pathlib import Path
from typing import Dict, Set

from openai import OpenAI
from pylogics.syntax.base import Formula

from nl2ltl.engines.base import Engine
from nl2ltl.engines.gpt import ENGINE_ROOT
from nl2ltl.engines.gpt.output import GPTOutput, parse_gpt_output, parse_gpt_result
from nl2ltl.filters.base import Filter

try:
    client = OpenAI()
except Exception:
    client = None

engine_root = ENGINE_ROOT
DATA_DIR = engine_root / "data"
PROMPT_PATH = engine_root / DATA_DIR / "prompt.json"


class Models(Enum):
    """The set of available GPT language models."""

    GPT35_INSTRUCT = "gpt-3.5-turbo-instruct"
    GPT35_TURBO = "gpt-3.5-turbo"
    GPT4 = "gpt-4"


SUPPORTED_MODELS: Set[str] = {v.value for v in Models}


class OperationModes(Enum):
    """The set of available GPT modes."""

    CHAT = "chatCompletion"
    COMPLETION = "Completion"


SUPPORTED_MODES: Set[str] = {v.value for v in OperationModes}


class GPTEngine(Engine):
    """The GPT engine."""

    def __init__(
        self,
        model: str = Models.GPT35_TURBO.value,
        prompt: Path = PROMPT_PATH,
        operation_mode: str = OperationModes.CHAT.value,
        temperature: float = 0.5,
    ):
        """GPT LLM Engine initialization."""
        self._model = model
        self._prompt = self._load_prompt(prompt)
        self._operation_mode = operation_mode
        self._temperature = temperature

        self._check_consistency()

    def _load_prompt(self, prompt):
        return json.load(open(prompt))["prompt"]

    def _check_consistency(self) -> None:
        """Run consistency checks."""
        self.__check_openai_version()
        self.__check_model_support()
        self.__check_operation_mode()

    def __check_openai_version(self):
        """Check that the GPT tool is at the right version."""
        is_right_version = client._version == "1.12.0"
        if not is_right_version:
            raise Exception(
                "OpenAI needs to be at version 1.12.0. "
                "Please install it manually using:"
                "\n"
                "pip install openai==1.12.0"
            )

    def __check_model_support(self):
        """Check if the model is a supported model."""
        is_supported = self.model in SUPPORTED_MODELS
        if not is_supported:
            raise Exception(f"The LLM model {self.model} is not currently supported by nl2ltl.")

    def __check_operation_mode(self):
        """Check if the operation mode is a supported mode."""
        is_supported = self.operation_mode in SUPPORTED_MODES
        if not is_supported:
            raise Exception(f"The operation mode {self.operation_mode} is not currently supported by nl2ltl.")

    @property
    def model(self) -> str:
        """Get the GPT model."""
        return self._model

    @property
    def prompt(self) -> str:
        """Get the GPT prompt."""
        return self._prompt

    @property
    def operation_mode(self) -> str:
        """Get the GPT operation mode."""
        return self._operation_mode

    @property
    def temperature(self) -> float:
        """Get the GPT temperature."""
        return self._temperature

    def translate(self, utterance: str, filtering: Filter = None) -> Dict[Formula, float]:
        """From NL to best matching LTL formulas with confidence."""
        return _process_utterance(
            utterance,
            self.model,
            self.prompt,
            self.operation_mode,
            self.temperature,
            filtering,
        )


def _process_utterance(
    utterance: str,
    model: str,
    prompt: str,
    operation_mode: str,
    temperature: float,
    filtering: Filter,
) -> Dict[Formula, float]:
    """Process NL utterance.

    :param utterance: the natural language utterance
    :param model: the GPT model
    :param prompt: the prompt
    :param operation_mode: the operation mode
    :param temperature: the temperature
    :param filtering: the filter used to remove formulas
    :return: a dict matching formulas to their confidence
    """
    query = f"NL: {utterance}\n"
    messages = [{"role": "user", "content": prompt + query}]
    if operation_mode == OperationModes.CHAT.value:
        prediction = client.chat.completions.create(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=200,
            top_p=1.0,
            frequency_penalty=0.0,
            presence_penalty=0.0,
            stop=["\n\n"],
        )
    else:
        prediction = client.completions.create(
            model=model,
            prompt=messages[0]["content"],
            temperature=temperature,
            max_tokens=200,
            top_p=1.0,
            frequency_penalty=0.0,
            presence_penalty=0.0,
            stop=["\n\n"],
        )

    gpt_result: GPTOutput = parse_gpt_output(prediction, operation_mode)
    matching_formulas: Dict[Formula, float] = parse_gpt_result(gpt_result, filtering)
    return matching_formulas



"""This module implements the library exceptions."""


class Nl2ltlfException(Exception):
    """Base library exception."""


class NotImplementedEngineFunction(Nl2ltlfException):
    """Raise this exception when an engine does not support an operation."""


class BadUtteranceException(Nl2ltlfException):
    """Raise this exception when the utterance is malformed."""

    __ERROR_MSG = "wrong utterance for method {method_name}: expected '{expected}', found '{actual}'"

    def __init__(self, method_name: str, expected: str, actual: str, *args, **kwargs):
        """Initialize the exception."""
        format_args = dict(method_name=method_name, expected=expected, actual=actual)
        super().__init__(self.__ERROR_MSG.format(**format_args))
